
--------------------------------------------------------------------------------
--				     TKSESH
--                          Ancient Egyptian Database
--
-- 
--		      Table creation for SQL Lite Database
-- 
-- headers are the only real identifiers for an entry.  for internal use,
-- simpler numeric identifiers will be used, but headers are always used for
-- data exchange (even data dumps should use headers)

create table Header (
       -- Header : common to all signed data
       -- (inheritance is the key word here)
       -- creator
       uid integer,
       -- creator machine number
       mid integer,
       -- specific id for this item
       id integer,
       -- type of this entry.
       type varchar(10),
       -- all signifiant entries have a date field.
       date date,
       -- the session id is used for two things :
       -- first, it allows one to group data
       -- second, a session number of "0" means that
       -- data has been "published" and should not be
       -- modified anymore.
       sid integer,
       -- the internal id which corresponds to this header
       repOid integer,
       primary key (uid,mid,id,type)
);

-- a glyph created by a user.

CREATE TABLE Glyph (
       oid integer,
       -- oid integer, automatically created by SQLITE
       -- A sign code is : UG<uid>m<mid>s<id>
       -- for instance, the first sign user 5 will create will have code :
       -- UG5m0s0
       -- Glyph-specific code
       -- id of the blob which contains the drawing
       drawing text, -- vectorial drawing of the Glyph
       guides text, -- guidelines for this glyph
       PRIMARY KEY (oid)
);

-- Identification of Glyphs with MDC codes.
-- Note that any glyph has automatically a corresponding,
-- *computed* MDC code of the form of U0m0s0
-- a glyph may be assigned different codes by different persons.
-- resp. a code may correspond to different glyphs

create table HasCode (
       -- This is an identifier for this entry
       oid integer, -- automatical with SQLITE
       -- the glyph :
       glyphId integer,
       -- the code
       code varchar(20),
       -- is the code the official one for this glyph ?
       isOfficial integer,
       PRIMARY KEY (oid)
);



-- Abstract Sign 
-- 	A sign identified by a code in the manuel.
--	For each Abstract Sign, we have 0 or 1 official
--	Glyph. 
--


create table AbstractSign (
       code varchar(20),
       glyphId integer, -- official Glyph. Currently not used.
       PRIMARY KEY (code)
);


-- PhonCode 
-- phonetic code in the manuel de codage.

create table PhonCode (
       code varchar(20) primary key,
       sign varchar(20) -- sign is an AbstractSign.
);

-- Associate an abstract sign with a possible value.

create table HasValue (
       oid integer primary key,
       code varchar(20), -- The abstract sign code
       -- Normally, we should build a value object, and HasValue
       -- should link an abstract sign and a value.
       -- currently, let's keep the value in the HasValue.
       -- valueID integer, NOT USED
       value varchar(20),
       -- type for the value
       valueType varchar(10),
       
       -- the user can give a 
       -- number of hints:

       -- is this value typical of
       -- the sign?
       -- rating from 0 to 5.
       -- 5 always
       -- 4 almost always
       -- 3 usual
       -- 2 not usual
       -- 1 rare
       -- 0 maybe an error
       usualValue integer,

       -- is the sign typical for this value ?
       -- same scale as usualValue.
       -- a nice possible use is for retrieving
       -- a sign for a given value

       usualSign integer,

       -- This will be a code for a period.
       period integer
);

-- 
-- Sign values.
-- a sign value can be many things.

-- To simplify coding, we represent the multiple 
-- sub-classes of SignValue in one table only :

-- create table SignValue (
--        -- identifier
--        oid integer PRIMARY KEY,
--        -- type of value.
--        -- 
--        -- 0 means "phonogram"
--        type integer
-- );

-- create table Phonogram (
--       -- identifier. Should be the same
--       -- as the corresponding SignValue
--       -- that's inheritance built in relational 
--       -- databases
--       oid integer primary key,
--       -- the value
--       value varchar(20),
--       -- specific classifier.
--       -- 
--       subtype integer
-- );


-- is a sign related to another sign ?

create table SignRelation (
       oid integer PRIMARY KEY,
       code1 varchar(20),
       code2 varchar(20),
       -- type of relation :
       -- variant of : "v"
       -- mutilated variant of : "m"
       -- hieratic variant of : "h"
       -- confused with : "c"
       -- part of : "p"
       reltype char,
       -- confidence on a scale of 0 to 5.

       confidence integer
       -- PRIMARY KEY (oid)	
);

-- -- this value indicates that a sign is confused with another one.

-- create table Confusion (
--        oid integer PRIMARY KEY,
--        code varchar(20),
-- );


-- Multi-lingual system.

-- a MLDesignation stands for "something that has a name"

create table MLDesignation (
       -- unique word or expression ID
       oid integer PRIMARY KEY,
       defaultString varchar(255)
);

-- actual names for MLDesignations.
-- names can be given by users.
-- multiple users can have
-- more than one designation
-- for the same thing in the same language.
--
-- possibilities are :
--
-- let's say, text names :
--	 "Le paysan loquent"
--	 "L'oasien"
--	 "The eloquent Peasant"
--	 "The Peasant"
--	 and so and so...
--
-- 

create table MLStrings (
       oid integer primary key,
       -- the language is coded according
       -- to ....
       lang char(2),
       text varchar(255),
       mlCode integer -- multilingual entry
);



--------------------------------------------------------------------------
-- Text oriented part
--
-- very coarse text system by now.

create table Text (
	oid integer primary key,
	-- mlOid integer, -- MLDesignations for this text
	name text, -- designation for this text. Should not make it in the final system
	refTextOid, -- for translitteration and translation,
	            -- the text which is the basis of the work.
		    -- If refTextOid is not NULL, then name should be NULL. 
	comment text, -- a creator comment for this text.
	textType text, -- the text type. "hie" stands for hieroglyphic.
	               -- "tri" stands for transl*i*tteration, 
		       -- and "tra" for transl*a*tion
	body text  -- the body, as UTF-8 text
);

-- secondary text, *automatically* built from a text. E.G. list of signs in a text.

create table SubText (
	textOid integer,
	subtype text,
	subbody text,
	primary key (textOid, subtype)
);

-- a text translation, in a simple format :
-- a hieroglyphic text, a translitteration, and a translation

create table SimpleTranslation (
	oid integer primary key,
	hieroglyphicTextOid integer,
	translitterationTextOid integer,
	translationTextOid integer
);

-- Final system. Not used yet.
--
-- A TextWork is a text analysis, with translations, translitterations, comments...
-- The system supports parallel texts, with one or multiple translations.

create table TextWork (
	oid integer primary key,
	title text,
	first integer, -- the first segment's oid.
	last integer -- the last segment's oid.
);

-- 
-- type is either "r" if the segment is directly contained in a TextWork,
-- or "s" if the segment is contained in another segment.
--
-- The meaning of previous and next change in these cases.
--
-- For "r" segment, previous and next refer to the text sequence.
-- That is, a segment b is the next of segment a if b comes just after a in the text.
--
-- for "s" segments, previous and next deals more with an order of presentation
-- for text items which are essentially parallel; for instance, a translation might be the "next"-- of a translitteration.

create table TextWorkSegment (
	oid integer primary key,
	type text,
	previous integer,
	next integer
);


-- Text Work Items should definitly be objects. One can comment them, for instance.
create table TextWorkItem (
	oid integer primary key,
	parent integer, 
	type integer, -- 'r' for reference, 'h' for hieroglyphs, 
	              -- 't' for translitteration, 'a' for translation, 'c' for comment.
	content text, -- content. Manuel de codage or plain text.
	textOid integer, -- For references.
	start text, -- start position in the reference text.
	end text, -- ending position in the reference text.
	-- The following line is here just as a demonstration of sqlite's parser.
	-- It doesn't do anything
	foreign key (parent) references TextWorkSegment
);

-- For later use.
create table TextReference (
	textOid integer,
	start text, -- start position in the reference text.
	end text, -- ending position in the reference text.
	primary key (textOid, start, end)
);

/**
**	A note about texts naming.
**	Texts can be designated in a number of ways. When looking at actual instances of texts,
**	one sees that none of these ways is sufficient by itself. Possibilities are :
**
**	* a text can be designated by its edition (e.g. position in the LRL or KRI, etc...)
**	  some papyri, for instance, have no inventory number and can only be distinguished
**	  by their edition
**
**	* a text can be designated by its primary source. e.g. "O. DM. 123"
**
**	* a text can be designated by one of the names of the composition it contains.
**	  e.g. "The Eloquent Peasant". Note that this kind of
** 	  references is often used alongside another one. For instance "Peasant, B 124".
**
**/

--------------------------------------------------------------------------
-- Dictionnary 
--
-- A dictionnary is a graph

--create table DictionnaryNode (
--);

-- TEMPORARY SYSTEM FOR DICTIONARY
-- This system is simply a port of the old tksesh dictionary to sqlite.
-- IT WILL CHANGE IN THE NEXT TKSESH Version.

-- Entry in the dictionary :

create table TmpDico (
	oid integer,
	entry text,
	PRIMARY KEY (oid)
);

-- Index for entries

-- Notice that we have two "values" for a tag.
-- The first one is the lookup value, and the second the actual value.
-- 
-- The idea is that we usually want 
-- 	* to normalize a value before searching (for instance, to drop points in translitteration)
--	* to order the entries in a way that can be quite different from the ascii order.
--	  e.g. we want translitterations to be ordered according to the standard egyptological alphabetic order,
--	  and not according to the ASCII order.

create table tmpDicoIdx (
	tag varchar(10), -- the tag's name
	lookup text, -- the value used for searching and sorting.
	value text, -- the tag's real value
	ref integer -- an entry which contains this tag (internal object reference !)
);

-- orthographies
create table tmpOrtho (
	translit text, -- a word translitteration (normalized)
	ortho text -- a written form for this word, in MdC encoding.
);


--------------------------------------------------------------------------
--	Tables for text edition and translation.
-- 	
--      


--------------------------------------------------------------------------
--		  		Technical tables :
-- 	These tables are used internally. They are not supposed to be seen
--      from outside.

create table ClassName (
	classname varchar(10) primary key
);	

insert into ClassName values ('Glyph');
insert into ClassName values ('HasCode');
insert into ClassName values ('HasValue');
insert into ClassName values ('SignRelation');
insert into ClassName values ('MLDesignation');
insert into ClassName values ('MLStrings');
insert into ClassName values ('Text');
insert into ClassName values ('TmpDico');
insert into ClassName values ('SimpleTranslation');


-- Management table :
-- this database's user
-- user's next session id.
-- normally, this table has only got one row, for this database user and this
-- machine. The structure would allow sharing in a network environment. 


create table Management (
	uid integer,
       	mid integer,
	name text, -- user name
	mail text, -- user mail
       	maxsid integer
);

-- MaxTable : used to record :
-- the next available id for an entry in a given table
-- the last internal OID used for an entry (in order to link the entry to its header)

create table MaxTable (
	type varchar(10),
	maxid integer,
	lastOid integer, -- the last Oid given to a specific part for this kind of objects.
	primary key (type)
);


----------------------------------------------------------------------
--				Views
--
-- the "view" mecanism of sqlite is quite usable now. We create 
-- modifiable view which maintains most needs
--
-- typically, these views should be used when updating or modifying data.
--
-- a note on insertion in views : 
-- 	After giving some thought to the problem, we have decided 
-- 

-- view a glyph with one of its codes.
-- the code is here to ease glyph selection, and 
-- not for modification.

create view GlyphView as
 		select uid, mid, id, repOid, date, sid,
		       drawing, guides
		from Header, Glyph
		where Header.type = 'Glyph'
		and Glyph.oid = Header.repOid;

create trigger insertGlyphView instead of insert on GlyphView
	for each row 
	begin
		-- create the glyph entry
		insert into Glyph values (NULL, NEW.drawing, NEW.guides);

		update MaxTable 
		set lastOid= last_insert_rowid()
		where type = 'Glyph';

		-- create the header entry
		insert into HeaderUtil 
		select NEW.uid, NEW.mid, NEW.id, 'Glyph', NEW.date, NEW.sid, MaxTable.lastOid
		from MaxTable where type= 'Glyph';

		-- create the code entry
		insert into HasCodeView 
		select
			NEW.uid, NEW.mid, NULL, NULL, NEW.date,NEW.sid, 
			MaxTable.lastOid,
			'UG' || Header.uid || 'M' || Header.mid || 'N' || Header.id,
			1
		from MaxTable, Header
		where Header.repOid= MaxTable.lastOid 
		and Header.type = 'Glyph'
		and MaxTable.type = 'Glyph';
	end;

create trigger deleteGlyphView instead of delete on GlyphView
	for each row 
	begin
		-- delete the associated data :
		delete from Header 
		where uid = OLD.uid and mid = OLD.mid and id = OLD.id and type= 'Glyph';

		delete from Glyph where oid = OLD.repOid;

		-- delete the glyph abstract name ?
		-- maybe not a good idea ?

		delete from AbstractSign 
		where code = 'UG' || OLD.uid || 'M' || OLD.mid || 'N' || OLD.id;

		-- delete the glyph code associations :
	
		delete from HasCodeView where glyphId = OLD.repOid;

		-- delete all values for this sign... (that's something for AbstractSign).
	end;

create trigger updateGlyphView instead of update on GlyphView
	for each row 
	begin
		update Header
		set 
			date = NEW.date,
			sid = NEW.sid
		where uid = OLD.uid and mid = OLD.mid and id = OLD.id and type = 'Glyph';
		
		update Glyph
		set
			drawing = NEW.drawing,
			guides = NEW.guides
		where
			oid = OLD.repOid
		;
		
		update MaxTable
		set
			lastOid = NEW.repOid
		where
			type= 'Glyph'
		;
	end;


create view HasCodeView as 
 		select uid, mid, id, repOid, date, sid,
		       glyphId, code, isOfficial
		from Header, HasCode
		where	Header.type = 'HasCode'
		and 	HasCode.oid = Header.repOid;

create trigger insertHasCodeView instead of insert on HasCodeView
	for each row
	begin
		-- create the hasCode entry
		insert into HasCode values (NULL, NEW.glyphId, NEW.code, NEW.isOfficial);

		-- save the OID 
		update MaxTable 
		set lastOid= last_insert_rowid()
		where type = 'HasCode';

		-- create the header entry
		insert into HeaderUtil 
		select NEW.uid, NEW.mid, NEW.id, 'HasCode', 
		       NEW.date, NEW.sid, lastOid
		from MaxTable
		where type = 'HasCode';

		-- create the AbstractSign if it doesn't exist.
		insert or replace into AbstractSign values (NEW.code, NULL);
	end;

create trigger updateHasCodeView instead of update on HasCodeView
	for each row
	begin
		update Header
		set 
			date = NEW.date,
			sid = NEW.sid
		where uid = OLD.uid and mid = OLD.mid and id = OLD.id and type= 'HasCode';
		
		update HasCode
		set
			glyphId = NEW.glyphId,
			code = NEW.code,
			isOfficial = NEW.isOfficial
		where
			oid = OLD.repOid
		;
	end;

create trigger deleteHasCodeView instead of delete on HasCodeView 
	for each row
	begin
		-- delete the associated data :
		delete from Header 
		where uid = OLD.uid and mid = OLD.mid and id = OLD.id and type= 'HasCode';

		delete from HasCode where oid = OLD.repOid;
	end;

create view HeaderUtil as select uid, mid, id, type, date, sid, repOid from Header;

create trigger makeHeader instead of insert on HeaderUtil
for each row when NEW.id ISNULL
begin
		insert into Header (uid,mid,id,type,date,sid,repOid)
		select
			NEW.uid, 
			NEW.mid,
			MaxTable.maxid + 1,
			NEW.type,
			NEW.date,
			NEW.sid,
			NEW.repOid
		from MaxTable where type = NEW.type;

		update MaxTable 
		set maxid = maxid + 1
		where type = NEW.type;
end;

-- Insertion stuff used when adding a Header with an ID.
-- Gets sure MaxTable is up to date

create trigger makeHeader1 instead of insert on HeaderUtil
for each row when NEW.id NOTNULL
begin
		insert into Header (uid,mid,id,type,date,sid,repOid)
		values (NEW.uid, NEW.mid, NEW.id, NEW.type, NEW.date, NEW.sid, NEW.repOid);

		update MaxTable 
		set maxid = NEW.id
		where type = NEW.type 
		and NEW.id > maxid 
		and NEW.uid in (select uid from Management);
end;

create view HasValueView as
		select uid, mid, id, repOid, date, sid,
			code, value, 
			valueType, 
			usualValue,
			usualSign, period
		from Header, HasValue
		where	Header.type= 'HasValue'
		and	HasValue.oid= Header.repOid;

create trigger insertHasValueView instead of insert on HasValueView
	for each row
	begin
		-- create the HasValue entry
		insert into HasValue 
		values (NULL,NEW.code, NEW.value, NEW.valueType,
			NEW.usualValue, NEW.usualSign, NEW.period);
		-- save the OID
		update MaxTable 
		set lastOid= last_insert_rowid()
		where type = 'HasValue';
	
		-- create the header entry
		insert into HeaderUtil
		select NEW.uid, NEW.mid, NEW.id, 'HasValue', 
		       NEW.date, NEW.sid, lastOid
		from MaxTable
		where type = 'HasValue';
	end;

create trigger updateHasValueView instead of update on HasValueView
	for each row
	begin
		update Header
		set 
			date = NEW.date,
			sid = NEW.sid
		where uid = OLD.uid and mid = OLD.mid and id = OLD.id and type= 'HasValue';

		update HasValue
		set
			code= NEW.code,
			value= NEW.value,
			valueType= NEW.valueType,
			usualValue= NEW.usualValue,
			usualSign= NEW.usualSign,
			period= NEW.period
		where
			oid= OLD.repOid
		;
	end;

create trigger deleteHasValueView instead of delete on HasValueView
	for each row
	begin
		delete from Header
		where uid = OLD.uid and mid = OLD.mid and id = OLD.id and type= 'HasValue';
		
		delete from HasValue where oid= OLD.repOid;
	end;


create view SignRelationView as
	select uid, mid, id, repOid, date, sid,
		code1, code2, reltype, confidence
	from Header SignRelation
	where Header.type = 'SignRelation'
	and SignRelation.oid = Header.repOid;

create trigger insertSignRelation instead of insert on SignRelationView
	for each row 
	begin
		-- create the SignRelation
		insert into SignRelation values (NULL, NEW.code1, NEW.code2, 
						 NEW.reltype, NEW.confidence);
		
		-- bookkeeping to indentify the entry
		update MaxTable
		set lastOid= last_insert_rowid()
		where type = 'SignRelation';
		
		-- create the header entry
		insert into HeaderUtil 
		select NEW.uid, NEW.mid, NEW.id, 'SignRelation', 
		       NEW.date, NEW.sid, MaxTable.lastOid
		from MaxTable where type= 'SignRelation';
	end;

create trigger deleteSignRelationView instead of delete on SignRelationView
	for each row 
	begin
		-- delete the associated data :
		delete from Header 
		where uid = OLD.uid and mid = OLD.mid and id = OLD.id and type= 'SignRelation';
	
		delete from SignRelation where oid= OLD.repOid;
	end;

create trigger updateSignRelationView instead of update on SignRelationView
	for each row
	begin
		update Header
		set 
			date = NEW.date,
			sid = NEW.sid
		where uid = OLD.uid and mid = OLD.mid and id = OLD.id and type = 'SignRelation';
	
		update SignRelation
		set
			code1= NEW.code1,
			code2= NEW.code2,
			reltype= NEW.reltype,
			confidence= NEW.confidence
		where
			oid= OLD.repOid;

		update MaxTable
		set
			lastOid = NEW.repOid
		where
			type= 'SignRelation'
		;
	end;

create view TextView as
 		select uid, mid, id, repOid, date, sid,
		      name, refTextOid, comment, textType, body
		from Header, Text
		where Header.type = 'Text'
		and Text.oid = Header.repOid;

create trigger insertTextView instead of insert on TextView 
	for each row 
	begin
		-- create the text entry
		insert into Text values (NULL, NEW.name, NEW.refTextOid, 
					       NEW.comment, NEW.textType, NEW.body);
		
		-- bookkeeping to indentify the entry
		update MaxTable
		set lastOid= last_insert_rowid()
		where type = 'Text';
		
		-- create the header entry
		insert into HeaderUtil 
		select NEW.uid, NEW.mid, NEW.id, 'Text', NEW.date, NEW.sid, MaxTable.lastOid
		from MaxTable where type= 'Text';
	end;

create trigger deleteTextView instead of delete on TextView
	for each row 
	begin
		-- delete the associated data :
		delete from Header 
		where uid = OLD.uid and mid = OLD.mid and id = OLD.id and type= 'Text';
	
		delete from Text where oid= OLD.repOid;

		delete from SubText where textOid= OLD.repOid;
	end;

create trigger updateTextView instead of update on TextView
	for each row
	begin
		update Header
		set 
			date = NEW.date,
			sid = NEW.sid
		where uid = OLD.uid and mid = OLD.mid and id = OLD.id and type = 'Text';
	
		update Text
		set
			name= NEW.name,
			refTextOid= NEW.refTextOid,
			comment= NEW.comment,
			textType= NEW.textType,
			body= NEW.body
		where
			oid= OLD.repOid;

		update MaxTable
		set
			lastOid = NEW.repOid
		where
			type= 'Text'
		;
	end;

--- Temporary view for TmpDico :
create view TmpDicoView as
	select uid, mid, id, repOid, date, sid,
		entry
	from Header, TmpDico
	where Header.type= 'TmpDico'
	and   TmpDico.oid= Header.repOid;

create trigger insertTmpDicoView instead of insert on TmpDicoView
	for each row
	begin
	-- create the TmpDico entry 
	insert into TmpDico values (NULL, NEW.entry);

	-- bookkeeping to indentify the entry
	update MaxTable
	set lastOid= last_insert_rowid()
	where type = 'TmpDico';

	-- create the header entry 
	insert into HeaderUtil 
	select NEW.uid, NEW.mid, NEW.id, 'TmpDico', NEW.date, NEW.sid, MaxTable.lastOid
	from MaxTable where type= 'TmpDico';
	end;

create trigger updateTmpDicoView instead of update on TmpDicoView
	for each row
	begin	
		update Header
		set 
			date = NEW.date,
			sid = NEW.sid
		where uid = OLD.uid and mid = OLD.mid and id = OLD.id and type = 'TmpDico';
		
		update TmpDico
		set
			entry = NEW.entry
		where
			oid= OLD.repOid;

		update MaxTable
		set
			lastOid = NEW.repOid
		where
			type= 'TmpDico'
		;
	end;

create trigger deleteTmpDicoView instead of delete on TmpDicoView
	for each row
	begin
		-- delete the associated data :
		delete from Header 
		where uid = OLD.uid and mid = OLD.mid and id = OLD.id and type= 'TmpDico';
	
		delete from TmpDico where oid= OLD.repOid;

		delete from TmpDicoIdx where ref = OLD.repOid;
	end;

-- insert into Management values (1,1,0,1,0);
-- insert into GlyphView values (NULL,NULL,NULL,NULL, 'today', 1, 'un dessin','');
-- insert into HasCodeView values (NULL,NULL,NULL,NULL,'today',1,100,'AAAA',0); 

-- indexes :
create index codeIndexOnHasCode on HasCode (code);
create index glyphIndexOnHasCode on HasCode (glyphId);
create index reverseHeaderIndex on Header (repOid);
create index codeIndexOnHasValue on HasValue (code);
create index valueIndexOnHasValue on HasValue (value);

--------------------------------------------------------------------------
-- Secondary tables, used for caching data.

create table SignBitmap (
	code varchar(10) primary key,
	bitmap text
);

-- We store ligature definitions in the database,
-- for the sake of saving everything there.
-- however, it should be clear that the ligature system
-- is *not* safe. See RES (from M.-J. Nederhof) for a consistant system.
-- In any way, ligatures are not considered as user entries.

create table ligature (
	ligCode varchar(30) primary key,
	ligRep text
);


-- The glossary is essentially personnal, used as a facility for a user, and 
-- does not store significant data.

create table glossary (
	code text primary key,
	value text
);
