#==============================================================================
# Demonstrates how to implement a tablelist widget for displaying the
# configuration options of an arbitrary widget.
#
# Copyright (c) 2000-2002  Csaba Nemethi (E-mail: csaba.nemethi@t-online.de)
#==============================================================================

package require Tablelist

namespace eval demo {
    #
    # Add some entries to the Tk option database for the following
    # widget hierarchy within a top-level widget of the class DemoTop:
    #
    # Name		Class
    # -----------------------------
    # tf		Frame
    #   tbl		  Tabellist
    #   vsb, hsb	  Scrollbar
    # bf		Frame
    #   b1, b2, b3	  Button
    #
    switch $::tcl_platform(platform) {
	unix {
	    option add *DemoTop*Font				"Helvetica -12"
	}
	windows {
	    option add *DemoTop.tf.borderWidth			2
	    option add *DemoTop.tf.relief			sunken
	    option add *DemoTop.tf.tbl.borderWidth		0
	    option add *DemoTop.tf.tbl.highlightThickness	0
	}
	macintosh {
	    option add *DemoTop*Background			#dedede

	    option add *DemoTop.tf.borderWidth			2
	    option add *DemoTop.tf.relief			sunken
	    option add *DemoTop.tf.tbl.borderWidth		0
	    option add *DemoTop.tf.tbl.highlightThickness	0
	}
    }
    option add *DemoTop.tf.tbl.background		gray96
    option add *DemoTop.tf.tbl.stripeBackground		#e0e8f0
    option add *DemoTop.tf.tbl.selectBackground		navy
    option add *DemoTop.tf.tbl.selectForeground		white
    option add *DemoTop.tf.tbl.setGrid			yes
    option add *DemoTop.bf.Button.width			10
}

#------------------------------------------------------------------------------
# demo::displayConfig
#
# Displays the configuration options of the widget w in a tablelist widget
# contained in a newly created top-level widget.  Returns the name of the
# tablelist widget.
#------------------------------------------------------------------------------
proc demo::displayConfig w {
    if {![winfo exists $w]} {
	bell
	tk_messageBox -icon error -message "Bad window path name \"$w\"" \
		      -type ok
	return ""
    }

    #
    # Create a top-level widget of the class DemoTop
    #
    set top .configTop
    for {set n 2} {[winfo exists $top]} {incr n} {
	set top .configTop$n
    }
    toplevel $top -class DemoTop
    wm title $top "Configuration Options of the [winfo class $w] Widget \"$w\""

    #
    # Create a scrolled tablelist widget with 5 dynamic-width
    # columns and interactive sort capability within the top-level
    #
    set tf $top.tf
    frame $tf
    set tbl $tf.tbl
    set vsb $tf.vsb
    set hsb $tf.hsb
    tablelist::tablelist $tbl \
	-columns {0 "Command-Line Name"
		  0 "Database/Alias Name"
		  0 "Database Class"
		  0 "Default Value"
		  0 "Current Value"} \
	-labelcommand tablelist::sortByColumn -sortcommand demo::compareAsSet \
	-xscrollcommand [list $hsb set] -yscrollcommand [list $vsb set] \
	-height 15 -width 100 -stretch all
    scrollbar $vsb -orient vertical   -command [list $tbl yview]
    scrollbar $hsb -orient horizontal -command [list $tbl xview]

    #
    # Create three buttons within a frame child of the top-level widget
    #
    set bf $top.bf
    frame $bf
    set b1 $bf.b1
    set b2 $bf.b2
    set b3 $bf.b3
    button $b1 -text "Refresh"     -command [list demo::putConfig $w $tbl]
    button $b2 -text "Sort as set" -command [list $tbl sort]
    button $b3 -text "Close"       -command [list destroy $top]

    #
    # Manage the widgets
    #
    grid $tbl -row 0 -column 0 -sticky news
    grid $vsb -row 0 -column 1 -sticky ns
    grid $hsb -row 1 -column 0 -sticky ew
    grid rowconfigure    $tf 0 -weight 1
    grid columnconfigure $tf 0 -weight 1
    pack $b1 $b2 $b3 -side left -expand yes
    pack $bf -side bottom -fill x -pady 10
    pack $tf -side top -expand yes -fill both

    #
    # Fill the tablelist with the configuration options of the given widget
    #
    putConfig $w $tbl
    return $tbl
}

#------------------------------------------------------------------------------
# demo::putConfig
#
# Outputs the configuration options of the widget w into the tablelist widget
# tbl.
#------------------------------------------------------------------------------
proc demo::putConfig {w tbl} {
    if {![winfo exists $w]} {
	bell
	tk_messageBox -icon error -message "Bad window path name \"$w\"" \
		      -parent [winfo toplevel $tbl] -type ok
	return ""
    }

    #
    # Display the configuration options of w in the tablelist widget tbl
    #
    $tbl delete 0 end
    foreach configSet [$w configure] {
	#
	# Enclose the default and current values between braces if necessary
	#
	if {[llength $configSet] != 2} {
	    foreach {opt dbName dbClass default current} $configSet {
		if {[llength $default] != 1} {
		    set default \{$default\}
		}
		if {[llength $current] != 1} {
		    set current \{$current\}
		}
	    }
	    set configSet {}
	    lappend configSet $opt $dbName $dbClass $default $current
	}

	#
	# Insert the list configSet into the tablelist widget
	#
	$tbl insert end $configSet

	#
	# Change the colors of the first and last cell of the row
	# if the current value is different from the default one
	#
	if {[llength $configSet] != 2 &&
	    [string compare $default $current] != 0} {
	    foreach col {0 4} {
		$tbl cellconfigure end,$col \
		     -foreground red -selectforeground yellow
	    }
	}
    }

    $tbl sortbycolumn 0
}

#------------------------------------------------------------------------------
# demo::compareAsSet
#
# Compares two items of a tablelist widget used to display the configuration
# options of an arbitrary widget.  The item in which the current value is
# different from the default one is considered to be less than the other; if
# both items fulfil this condition or its negation then string comparison is
# applied to the two option names.
#------------------------------------------------------------------------------
proc demo::compareAsSet {item1 item2} {
    foreach {opt1 dbName1 dbClass1 default1 current1} $item1 \
	    {opt2 dbName2 dbClass2 default2 current2} $item2 {
	set changed1 [expr {[string compare $default1 $current1] != 0}]
	set changed2 [expr {[string compare $default2 $current2] != 0}]
	if {$changed1 == $changed2} {
	    return [string compare $opt1 $opt2]
	} elseif {$changed1} {
	    return -1
	} else {
	    return 1
	}
    }
}

#------------------------------------------------------------------------------

return "\nTo display the configuration options of an arbitrary\
	widget, enter\n\n\tdemo::displayConfig <widgetName>\n"
